import React, { useCallback, useContext, useImperativeHandle, useMemo, useState } from "react";
import { StyleSheet } from "react-native";
import { PanGestureHandler } from "react-native-gesture-handler";
import Animated, { runOnJS, useAnimatedGestureHandler, useAnimatedStyle, useDerivedValue, useSharedValue, withSpring } from "react-native-reanimated";
export let OpenDirection;

(function (OpenDirection) {
  OpenDirection["LEFT"] = "left";
  OpenDirection["RIGHT"] = "right";
  OpenDirection["NONE"] = "none";
})(OpenDirection || (OpenDirection = {}));

const renderNull = () => null;

const UnderlayContext = /*#__PURE__*/React.createContext(undefined);
const OverlayContext = /*#__PURE__*/React.createContext(undefined);

function SwipeableItem(props, ref) {
  const {
    item,
    children,
    renderOverlay = renderNull,
    renderUnderlayLeft = renderNull,
    renderUnderlayRight = renderNull,
    snapPointsLeft = [],
    snapPointsRight = [],
    swipeEnabled,
    activationThreshold = 20,
    overSwipe = 20,
    swipeDamping = 10,
    onChange = () => {},
    animationConfig = {}
  } = props;
  const springConfig = {
    damping: 20,
    mass: 0.2,
    stiffness: 100,
    overshootClamping: false,
    restSpeedThreshold: 0.5,
    restDisplacementThreshold: 0.5,
    ...animationConfig
  };
  const [openDirection, setOpenDirection] = useState(OpenDirection.NONE);
  const animStatePos = useSharedValue(0);
  const isGestureActive = useSharedValue(false);
  const swipingLeft = useDerivedValue(() => animStatePos.value < 0, [animStatePos]);
  const swipingRight = useDerivedValue(() => animStatePos.value > 0, [animStatePos]);
  const maxSnapPointLeft = -1 * (Math.max(...snapPointsLeft) || 0);
  const maxSnapPointRight = Math.max(...snapPointsRight) || 0;
  const maxTranslateLeft = maxSnapPointLeft - overSwipe;
  const maxTranslateRight = maxSnapPointRight + overSwipe;
  const percentOpenLeft = useDerivedValue(() => {
    return swipingLeft.value ? Math.abs(animStatePos.value / maxSnapPointLeft) : 0;
  }, []);
  const percentOpenRight = useDerivedValue(() => {
    return swipingRight.value ? Math.abs(animStatePos.value / maxSnapPointRight) : 0;
  }, []);
  const hasLeft = !!(snapPointsLeft !== null && snapPointsLeft !== void 0 && snapPointsLeft.length);
  const hasRight = !!(snapPointsRight !== null && snapPointsRight !== void 0 && snapPointsRight.length);
  const activeOffsetL = hasLeft || openDirection === OpenDirection.RIGHT ? -activationThreshold : -Number.MAX_VALUE;
  const activeOffsetR = hasRight || openDirection === OpenDirection.LEFT ? activationThreshold : Number.MAX_VALUE;
  const activeOffsetX = [activeOffsetL, activeOffsetR];
  const leftStyle = useAnimatedStyle(() => ({
    opacity: percentOpenLeft.value > 0 ? 1 : 0
  }), []);
  const rightStyle = useAnimatedStyle(() => ({
    opacity: percentOpenRight.value > 0 ? 1 : 0
  }), []);
  const overlayStyle = useAnimatedStyle(() => ({
    transform: [{
      translateX: animStatePos.value
    }]
  }), [animStatePos]);

  function openLeft(snapPoint) {
    return new Promise(resolve => {
      function resolvePromiseIfFinished(isFinished) {
        if (isFinished) resolve();
      }

      animStatePos.value = withSpring(snapPoint !== null && snapPoint !== void 0 ? snapPoint : maxSnapPointLeft, springConfig, isFinished => {
        runOnJS(resolvePromiseIfFinished)(isFinished);
      });
    });
  }

  function openRight(snapPoint) {
    return new Promise(resolve => {
      function resolvePromiseIfFinished(isFinished) {
        if (isFinished) resolve();
      }

      animStatePos.value = withSpring(snapPoint !== null && snapPoint !== void 0 ? snapPoint : maxSnapPointRight, springConfig, isFinished => {
        runOnJS(resolvePromiseIfFinished)(isFinished);
      });
    });
  }

  function close() {
    return new Promise(resolve => {
      function resolvePromiseIfFinished(isFinished) {
        if (isFinished) resolve();
      }

      animStatePos.value = withSpring(0, springConfig, isFinished => {
        runOnJS(resolvePromiseIfFinished)(isFinished);
      });
    });
  }

  useImperativeHandle(ref, () => ({
    open: (openDirection, snapPoint) => {
      if (openDirection === OpenDirection.LEFT) return openLeft(snapPoint);
      if (openDirection === OpenDirection.RIGHT) return openRight(snapPoint);
      return close();
    },
    close
  }));

  function onAnimationEnd(openDirection, snapPoint) {
    setOpenDirection(openDirection);
    onChange({
      open: openDirection,
      snapPoint
    });
  }

  const onGestureEvent = useAnimatedGestureHandler({
    onStart: (evt, ctx) => {
      ctx.startX = animStatePos.value;
      isGestureActive.value = true;
    },
    onActive: (evt, ctx) => {
      // console.log("onActive!!", evt)
      const rawVal = evt.translationX + ctx.startX;
      const clampedVal = Math.min(Math.max(maxTranslateLeft, rawVal), maxTranslateRight);
      animStatePos.value = clampedVal;
    },
    onEnd: evt => {
      isGestureActive.value = false; // Approximate where item would end up with velocity taken into account

      const velocityModifiedPosition = animStatePos.value + evt.velocityX / swipeDamping;
      const allSnapPoints = snapPointsLeft.map(p => p * -1).concat(snapPointsRight); // The user is not required to designate [0] in their snap point array,
      // but we need to make sure 0 is a snap point.

      allSnapPoints.push(0);
      const closestSnapPoint = allSnapPoints.reduce((acc, cur) => {
        const diff = Math.abs(velocityModifiedPosition - cur);
        const prevDiff = Math.abs(velocityModifiedPosition - acc);
        return diff < prevDiff ? cur : acc;
      }, Infinity);

      const onComplete = () => {
        const openDirection = closestSnapPoint === 0 ? OpenDirection.NONE : closestSnapPoint > 0 ? OpenDirection.RIGHT : OpenDirection.LEFT;
        runOnJS(onAnimationEnd)(openDirection, Math.abs(closestSnapPoint));
      };

      if (animStatePos.value === closestSnapPoint) onComplete();else animStatePos.value = withSpring(closestSnapPoint, springConfig, onComplete);
    }
  });
  const sharedParams = useMemo(() => ({
    item,
    isGestureActive,
    close
  }), []);
  const underlayRightParams = useMemo(() => {
    return {
      open: openRight,
      percentOpen: percentOpenRight,
      direction: OpenDirection.RIGHT,
      ...sharedParams
    };
  }, [percentOpenRight, openRight, sharedParams]);
  const underlayLeftParams = useMemo(() => {
    return {
      open: openLeft,
      percentOpen: percentOpenLeft,
      direction: OpenDirection.LEFT,
      ...sharedParams
    };
  }, [item, percentOpenLeft, openLeft, sharedParams]);
  const overlayParams = useMemo(() => {
    // If there is only one swipe direction, use it as the 'open' function. Otherwise we need to choose one.
    const open = hasLeft && !hasRight ? openLeft : hasRight && !hasLeft ? openRight : openLeft;
    return {
      openLeft: openLeft,
      openRight: openRight,
      percentOpenLeft,
      percentOpenRight,
      openDirection,
      open,
      ...sharedParams
    };
  }, [openLeft, openRight, openDirection, percentOpenLeft, percentOpenRight, hasLeft, hasRight]);
  return /*#__PURE__*/React.createElement(OverlayContext.Provider, {
    value: overlayParams
  }, /*#__PURE__*/React.createElement(Animated.View, {
    pointerEvents: openDirection === OpenDirection.LEFT ? "auto" : "none",
    style: [styles.underlay, leftStyle]
  }, /*#__PURE__*/React.createElement(UnderlayContext.Provider, {
    value: underlayLeftParams
  }, renderUnderlayLeft(underlayLeftParams))), /*#__PURE__*/React.createElement(Animated.View, {
    pointerEvents: openDirection === OpenDirection.RIGHT ? "auto" : "none",
    style: [styles.underlay, rightStyle]
  }, /*#__PURE__*/React.createElement(UnderlayContext.Provider, {
    value: underlayRightParams
  }, renderUnderlayRight(underlayRightParams))), /*#__PURE__*/React.createElement(PanGestureHandler, {
    enabled: swipeEnabled,
    activeOffsetX: activeOffsetX,
    onGestureEvent: onGestureEvent
  }, /*#__PURE__*/React.createElement(Animated.View, {
    style: [styles.flex, overlayStyle]
  }, children, renderOverlay(overlayParams))));
}

export default /*#__PURE__*/React.forwardRef(SwipeableItem);
export function useUnderlayParams() {
  const underlayContext = useContext(UnderlayContext);

  if (!underlayContext) {
    throw new Error("useUnderlayParams must be called from within an UnderlayContext.Provider!");
  }

  return underlayContext;
}
export function useOverlayParams() {
  const overlayContext = useContext(OverlayContext);

  if (!overlayContext) {
    throw new Error("useOverlayParams must be called from within an OverlayContext.Provider!");
  }

  return overlayContext;
}
export function useSwipeableItemParams() {
  const overlayContext = useContext(OverlayContext);

  if (!overlayContext) {
    throw new Error("useOverlayParams must be called from within an OverlayContext.Provider!");
  }

  const underlayContext = useContext(UnderlayContext);
  const contextDirection = underlayContext === null || underlayContext === void 0 ? void 0 : underlayContext.direction;
  const open = useCallback((snapPoint, direction) => {
    const openFnLeft = overlayContext.openLeft;
    const openFnRight = overlayContext.openRight;
    const openDirection = direction || contextDirection;
    const openFn = openDirection === OpenDirection.LEFT ? openFnLeft : openFnRight;
    return openFn(snapPoint);
  }, [overlayContext, contextDirection]);
  const percentOpen = useMemo(() => {
    if (contextDirection) {
      // If we're calling from within an underlay context, return the open percentage of that underlay
      return contextDirection === OpenDirection.LEFT ? overlayContext.percentOpenLeft : overlayContext.percentOpenRight;
    } // Return the open percentage of the active swipe direction


    return overlayContext.openDirection === OpenDirection.LEFT ? overlayContext.percentOpenLeft : overlayContext.percentOpenRight;
  }, [overlayContext]);
  return { ...overlayContext,
    open,
    percentOpen
  };
}
const styles = StyleSheet.create({
  underlay: { ...StyleSheet.absoluteFillObject
  },
  flex: {
    flex: 1
  }
});
//# sourceMappingURL=index.js.map